@@ -7,14 +7,19 @@ import (
 	"time"
 
 	"github.com/golang/glog"
+	configv1 "github.com/openshift/api/config/v1"
+	cligoinformersv1 "github.com/openshift/client-go/config/informers/externalversions/config/v1"
+	cligolistersv1 "github.com/openshift/client-go/config/listers/config/v1"
+	"github.com/openshift/library-go/pkg/operator/v1helpers"
+	"github.com/openshift/machine-config-operator/internal"
 	mcfgv1 "github.com/openshift/machine-config-operator/pkg/apis/machineconfiguration.openshift.io/v1"
 	daemonconsts "github.com/openshift/machine-config-operator/pkg/daemon/constants"
 	mcfgclientset "github.com/openshift/machine-config-operator/pkg/generated/clientset/versioned"
 	"github.com/openshift/machine-config-operator/pkg/generated/clientset/versioned/scheme"
 	mcfginformersv1 "github.com/openshift/machine-config-operator/pkg/generated/informers/externalversions/machineconfiguration.openshift.io/v1"
 	mcfglistersv1 "github.com/openshift/machine-config-operator/pkg/generated/listers/machineconfiguration.openshift.io/v1"
+	goerrs "github.com/pkg/errors"
 	corev1 "k8s.io/api/core/v1"
-	v1 "k8s.io/api/core/v1"
 	"k8s.io/apimachinery/pkg/api/errors"
 	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
 	"k8s.io/apimachinery/pkg/labels"
@@ -44,10 +49,16 @@ const (
 	// updateDelay is a pause to deal with churn in MachineConfigs; see
 	// https://github.com/openshift/machine-config-operator/issues/301
 	updateDelay = 5 * time.Second
-)
 
-// controllerKind contains the schema.GroupVersionKind for this controller type.
-var controllerKind = mcfgv1.SchemeGroupVersion.WithKind("MachineConfigPool")
+	// masterLabel defines the label associated with master node. The master taint uses the same label as taint's key
+	masterLabel = "node-role.kubernetes.io/master"
+
+	// workerLabel defines the label associated with worker node.
+	workerLabel = "node-role.kubernetes.io/worker"
+
+	// schedulerCRName that we're interested in watching.
+	schedulerCRName = "cluster"
+)
 
 var nodeUpdateBackoff = wait.Backoff{
 	Steps:    5,
@@ -70,6 +81,9 @@ type Controller struct {
 	mcpListerSynced  cache.InformerSynced
 	nodeListerSynced cache.InformerSynced
 
+	schedulerList         cligolistersv1.SchedulerLister
+	schedulerListerSynced cache.InformerSynced
+
 	queue workqueue.RateLimitingInterface
 }
 
@@ -77,6 +91,7 @@ type Controller struct {
 func New(
 	mcpInformer mcfginformersv1.MachineConfigPoolInformer,
 	nodeInformer coreinformersv1.NodeInformer,
+	schedulerInformer cligoinformersv1.SchedulerInformer,
 	kubeClient clientset.Interface,
 	mcfgClient mcfgclientset.Interface,
 ) *Controller {
@@ -87,7 +102,7 @@ func New(
 	ctrl := &Controller{
 		client:        mcfgClient,
 		kubeClient:    kubeClient,
-		eventRecorder: eventBroadcaster.NewRecorder(scheme.Scheme, v1.EventSource{Component: "machineconfigcontroller-nodecontroller"}),
+		eventRecorder: eventBroadcaster.NewRecorder(scheme.Scheme, corev1.EventSource{Component: "machineconfigcontroller-nodecontroller"}),
 		queue:         workqueue.NewNamedRateLimitingQueue(workqueue.DefaultControllerRateLimiter(), "machineconfigcontroller-nodecontroller"),
 	}
 
@@ -101,7 +116,11 @@ func New(
 		UpdateFunc: ctrl.updateNode,
 		DeleteFunc: ctrl.deleteNode,
 	})
-
+	schedulerInformer.Informer().AddEventHandler(cache.ResourceEventHandlerFuncs{
+		AddFunc:    ctrl.checkMasterNodesOnAdd,
+		UpdateFunc: ctrl.checkMasterNodesOnUpdate,
+		DeleteFunc: ctrl.checkMasterNodesOnDelete,
+	})
 	ctrl.syncHandler = ctrl.syncMachineConfigPool
 	ctrl.enqueueMachineConfigPool = ctrl.enqueueDefault
 
@@ -110,6 +129,9 @@ func New(
 	ctrl.mcpListerSynced = mcpInformer.Informer().HasSynced
 	ctrl.nodeListerSynced = nodeInformer.Informer().HasSynced
 
+	ctrl.schedulerList = schedulerInformer.Lister()
+	ctrl.schedulerListerSynced = schedulerInformer.Informer().HasSynced
+
 	return ctrl
 }
 
@@ -118,7 +140,7 @@ func (ctrl *Controller) Run(workers int, stopCh <-chan struct{}) {
 	defer utilruntime.HandleCrash()
 	defer ctrl.queue.ShutDown()
 
-	if !cache.WaitForCacheSync(stopCh, ctrl.mcpListerSynced, ctrl.nodeListerSynced) {
+	if !cache.WaitForCacheSync(stopCh, ctrl.mcpListerSynced, ctrl.nodeListerSynced, ctrl.schedulerListerSynced) {
 		return
 	}
 
@@ -132,11 +154,142 @@ func (ctrl *Controller) Run(workers int, stopCh <-chan struct{}) {
 	<-stopCh
 }
 
+func (ctrl *Controller) getCurrentMasters() ([]*corev1.Node, error) {
+	nodeList, err := ctrl.nodeLister.List(labels.SelectorFromSet(labels.Set{masterLabel: ""}))
+	if err != nil {
+		return nil, fmt.Errorf("error while listing master nodes %v", err)
+	}
+	return nodeList, nil
+}
+
+// checkMasterNodesOnAdd makes the master nodes schedulable/unschedulable whenever scheduler config CR with name
+// cluster is created
+func (ctrl *Controller) checkMasterNodesOnAdd(obj interface{}) {
+	ctrl.reconcileMasters()
+}
+
+// checkMasterNodesOnDelete makes the master nodes schedulable/unschedulable whenever scheduler config CR with name
+// cluster is created
+func (ctrl *Controller) checkMasterNodesOnDelete(obj interface{}) {
+	scheduler := obj.(*configv1.Scheduler)
+	if scheduler.Name != schedulerCRName {
+		glog.V(4).Infof("We don't care about CRs other than cluster created for scheduler config")
+		return
+	}
+	currentMasters, err := ctrl.getCurrentMasters()
+	if err != nil {
+		goerrs.Wrap(err, "Reconciling to make master nodes schedulable/unschedulable failed")
+		return
+	}
+	// On deletion make all masters unschedulable to restore default behaviour
+	errs := ctrl.makeMastersUnSchedulable(currentMasters)
+	if len(errs) > 0 {
+		err = v1helpers.NewMultiLineAggregate(errs)
+		goerrs.Wrap(err, "Reconciling to make nodes schedulable/unschedulable failed")
+		return
+	}
+	return
+}
+
+// checkMasterNodesonUpdate makes the master nodes schedulable/unschedulable whenever scheduler
+// config CR with name cluster is updated
+func (ctrl *Controller) checkMasterNodesOnUpdate(old, cur interface{}) {
+	oldScheduler := old.(*configv1.Scheduler)
+	curScheduler := cur.(*configv1.Scheduler)
+
+	if oldScheduler.Name != schedulerCRName || curScheduler.Name != schedulerCRName {
+		glog.V(4).Infof("We don't care about CRs other than cluster created for scheduler config")
+		return
+	}
+
+	if reflect.DeepEqual(oldScheduler.Spec, curScheduler.Spec) {
+		glog.V(4).Info("Scheduler config did not change")
+		return
+	}
+
+	ctrl.reconcileMasters()
+}
+
+// makeMastersUnSchedulable makes all the masters in the cluster unschedulable
+func (ctrl *Controller) makeMastersUnSchedulable(currentMasters []*corev1.Node) []error {
+	var errs []error
+	for _, node := range currentMasters {
+		if !CheckMasterIsAlreadySchedulable(node) {
+			continue
+		}
+		if err := ctrl.makeMasterNodeUnSchedulable(node); err != nil {
+			errs = append(errs, fmt.Errorf("failed making node %v schedulable with error %v", node.Name, err))
+		}
+	}
+	return errs
+
+}
+
+// makeMasterNodeUnSchedulable makes master node unschedulable by removing worker label and adding `NoSchedule`
+// master taint to the master node
+func (ctrl *Controller) makeMasterNodeUnSchedulable(node *corev1.Node) error {
+	_, err := internal.UpdateNodeRetry(ctrl.kubeClient.CoreV1().Nodes(), ctrl.nodeLister, node.Name, func(node *corev1.Node) {
+		// Remove worker label
+		newLabels := node.Labels
+		if _, hasWorkerLabel := newLabels[workerLabel]; hasWorkerLabel {
+			delete(newLabels, workerLabel)
+		}
+		node.Labels = newLabels
+		// Add master taint
+		newTaints := node.Spec.Taints
+		masterUnSchedulableTaint := corev1.Taint{Key: masterLabel, Effect: corev1.TaintEffectNoSchedule}
+		newTaints = append(newTaints, masterUnSchedulableTaint)
+		node.Spec.Taints = newTaints
+	})
+	if err != nil {
+		return err
+	}
+	return nil
+}
+
+// CheckMasterIsAlreadySchedulable checks if the given node has a worker label and doesn't have NoSchedule master
+// taint
+func CheckMasterIsAlreadySchedulable(node *corev1.Node) bool {
+	_, hasWorkerLabel := node.Labels[workerLabel]
+	hasMasterTaint := false
+	for _, taint := range node.Spec.Taints {
+		if taint.Key == masterLabel && taint.Effect == corev1.TaintEffectNoSchedule {
+			hasMasterTaint = true
+		}
+	}
+	return hasWorkerLabel && !hasMasterTaint
+}
+
+// makeMasterNodeSchedulable makes master node schedulable by removing NoSchedule master taint and
+// adding worker label
+func (ctrl *Controller) makeMasterNodeSchedulable(node *corev1.Node) error {
+	_, err := internal.UpdateNodeRetry(ctrl.kubeClient.CoreV1().Nodes(), ctrl.nodeLister, node.Name, func(node *corev1.Node) {
+		// Add worker label
+		newLabels := node.Labels
+		if _, hasWorkerLabel := newLabels[workerLabel]; !hasWorkerLabel {
+			newLabels[workerLabel] = ""
+		}
+		node.Labels = newLabels
+		// Remove master taint
+		newTaints := []corev1.Taint{}
+		for _, t := range node.Spec.Taints {
+			if t.Key == masterLabel && t.Effect == corev1.TaintEffectNoSchedule {
+				continue
+			}
+			newTaints = append(newTaints, t)
+		}
+		node.Spec.Taints = newTaints
+	})
+	if err != nil {
+		return err
+	}
+	return nil
+}
+
 func (ctrl *Controller) addMachineConfigPool(obj interface{}) {
 	pool := obj.(*mcfgv1.MachineConfigPool)
 	glog.V(4).Infof("Adding MachineConfigPool %s", pool.Name)
 	ctrl.enqueueMachineConfigPool(pool)
-
 }
 
 func (ctrl *Controller) updateMachineConfigPool(old, cur interface{}) {
@@ -165,6 +318,61 @@ func (ctrl *Controller) deleteMachineConfigPool(obj interface{}) {
 	// TODO(abhinavdahiya): handle deletes.
 }
 
+// Determine if masters are currently configured as schedulable
+func (ctrl *Controller) getMastersSchedulable() (bool, error) {
+	schedulerList, err := ctrl.schedulerList.List(labels.SelectorFromSet(nil))
+	if err != nil {
+		return false, fmt.Errorf("error while listing scheduler config %v", err)
+	}
+	for _, sched := range schedulerList {
+		if sched.Name == schedulerCRName {
+			return sched.Spec.MastersSchedulable, nil
+		}
+	}
+	return false, nil
+}
+
+// Determine if a given Node is a master
+func (ctrl *Controller) isMaster(node *corev1.Node) bool {
+	_, master := node.ObjectMeta.Labels[masterLabel]
+	return master
+}
+
+// Given a master Node, ensure it reflects the current mastersSchedulable setting
+func (ctrl *Controller) reconcileMaster(node *corev1.Node) {
+	mastersSchedulable, err := ctrl.getMastersSchedulable()
+	if err != nil {
+		goerrs.Wrap(err, "Getting scheduler config failed")
+		return
+	}
+	alreadySchedulable := CheckMasterIsAlreadySchedulable(node)
+	if mastersSchedulable && !alreadySchedulable {
+		err = ctrl.makeMasterNodeSchedulable(node)
+		if err != nil {
+			goerrs.Wrap(err, "Failed making master Node schedulable")
+			return
+		}
+	} else if !mastersSchedulable && alreadySchedulable {
+		err = ctrl.makeMasterNodeUnSchedulable(node)
+		if err != nil {
+			goerrs.Wrap(err, "Failed making master Node unschedulable")
+			return
+		}
+	}
+}
+
+// Get a list of current masters and apply scheduler config to them
+func (ctrl *Controller) reconcileMasters() {
+	currentMasters, err := ctrl.getCurrentMasters()
+	if err != nil {
+		goerrs.Wrap(err, "Reconciling to make master nodes schedulable/unschedulable failed")
+		return
+	}
+	for _, node := range currentMasters {
+		ctrl.reconcileMaster(node)
+	}
+}
+
 func (ctrl *Controller) addNode(obj interface{}) {
 	node := obj.(*corev1.Node)
 	if node.DeletionTimestamp != nil {
@@ -172,16 +380,22 @@ func (ctrl *Controller) addNode(obj interface{}) {
 		return
 	}
 
-	pool, err := ctrl.getPoolForNode(node)
+	if ctrl.isMaster(node) {
+		ctrl.reconcileMaster(node)
+	}
+
+	pools, err := ctrl.getPoolsForNode(node)
 	if err != nil {
-		glog.Errorf("error finding pools for node: %v", err)
+		glog.Errorf("error finding pools for node %s: %v", node.Name, err)
 		return
 	}
-	if pool == nil {
+	if pools == nil {
 		return
 	}
 	glog.V(4).Infof("Node %s added", node.Name)
-	ctrl.enqueueMachineConfigPool(pool)
+	for _, pool := range pools {
+		ctrl.enqueueMachineConfigPool(pool)
+	}
 }
 
 func (ctrl *Controller) updateNode(old, cur interface{}) {
@@ -192,7 +406,11 @@ func (ctrl *Controller) updateNode(old, cur interface{}) {
 		return
 	}
 
-	pool, err := ctrl.getPoolForNode(curNode)
+	if ctrl.isMaster(curNode) {
+		ctrl.reconcileMaster(curNode)
+	}
+
+	pool, err := ctrl.getPrimaryPoolForNode(curNode)
 	if err != nil {
 		glog.Errorf("error finding pool for node: %v", err)
 		return
@@ -205,17 +423,10 @@ func (ctrl *Controller) updateNode(old, cur interface{}) {
 	var changed bool
 	oldReadyErr := checkNodeReady(oldNode)
 	newReadyErr := checkNodeReady(curNode)
-	var oldReady, newReady string
-	if oldReadyErr != nil {
-		oldReady = oldReadyErr.Error()
-	} else {
-		oldReady = ""
-	}
-	if newReadyErr != nil {
-		newReady = newReadyErr.Error()
-	} else {
-		newReady = ""
-	}
+
+	oldReady := getErrorString(oldReadyErr)
+	newReady := getErrorString(newReadyErr)
+
 	if oldReady != newReady {
 		changed = true
 		if newReadyErr != nil {
@@ -231,20 +442,38 @@ func (ctrl *Controller) updateNode(old, cur interface{}) {
 		glog.Infof("Pool %s: node %s has completed update to %s", pool.Name, curNode.Name, curNode.Annotations[daemonconsts.DesiredMachineConfigAnnotationKey])
 		changed = true
 	} else {
-		annos := []string{daemonconsts.CurrentMachineConfigAnnotationKey, daemonconsts.DesiredMachineConfigAnnotationKey, daemonconsts.MachineConfigDaemonStateAnnotationKey}
+		annos := []string{
+			daemonconsts.CurrentMachineConfigAnnotationKey,
+			daemonconsts.DesiredMachineConfigAnnotationKey,
+			daemonconsts.MachineConfigDaemonStateAnnotationKey,
+		}
 		for _, anno := range annos {
 			if oldNode.Annotations[anno] != curNode.Annotations[anno] {
 				glog.Infof("Pool %s: node %s changed %s = %s", pool.Name, curNode.Name, anno, curNode.Annotations[anno])
 				changed = true
 			}
 		}
+		if !reflect.DeepEqual(oldNode.Labels, curNode.Labels) {
+			glog.Infof("Pool %s: node %s changed labels", pool.Name, curNode.Name)
+			changed = true
+		}
 	}
 
 	if !changed {
 		return
 	}
 
-	ctrl.enqueueMachineConfigPool(pool)
+	pools, err := ctrl.getPoolsForNode(curNode)
+	if err != nil {
+		glog.Errorf("error finding pools for node: %v", err)
+		return
+	}
+	if pools == nil {
+		return
+	}
+	for _, pool := range pools {
+		ctrl.enqueueMachineConfigPool(pool)
+	}
 }
 
 func (ctrl *Controller) deleteNode(obj interface{}) {
@@ -263,22 +492,24 @@ func (ctrl *Controller) deleteNode(obj interface{}) {
 		}
 	}
 
-	pool, err := ctrl.getPoolForNode(node)
+	pools, err := ctrl.getPoolsForNode(node)
 	if err != nil {
 		glog.Errorf("error finding pools for node: %v", err)
 		return
 	}
-	if pool == nil {
+	if pools == nil {
 		return
 	}
 	glog.V(4).Infof("Node %s delete", node.Name)
-	ctrl.enqueueMachineConfigPool(pool)
+	for _, pool := range pools {
+		ctrl.enqueueMachineConfigPool(pool)
+	}
 }
 
-// getPoolForNode chooses the MachineConfigPool that should be used for a given node.
+// getPoolsForNode chooses the MachineConfigPools that should be used for a given node.
 // It disambiguates in the case where e.g. a node has both master/worker roles applied,
-// and where a custom role may be used.
-func (ctrl *Controller) getPoolForNode(node *corev1.Node) (*mcfgv1.MachineConfigPool, error) {
+// and where a custom role may be used. It returns a slice of all the pools the node belongs to.
+func (ctrl *Controller) getPoolsForNode(node *corev1.Node) ([]*mcfgv1.MachineConfigPool, error) {
 	pl, err := ctrl.mcpLister.List(labels.Everything())
 	if err != nil {
 		return nil, err
@@ -324,16 +555,28 @@ func (ctrl *Controller) getPoolForNode(node *corev1.Node) (*mcfgv1.MachineConfig
 			return nil, fmt.Errorf("node %s has both master role and custom role %s", node.Name, custom[0].Name)
 		}
 		// One custom role, let's use its pool
-		return custom[0], nil
+		return []*mcfgv1.MachineConfigPool{custom[0], worker}, nil
 	} else if master != nil {
 		// In the case where a node is both master/worker, have it live under
 		// the master pool. This occurs in CodeReadyContainers and general
 		// "single node" deployments, which one may want to do for testing bare
 		// metal, etc.
-		return master, nil
+		return []*mcfgv1.MachineConfigPool{master}, nil
 	}
 	// Otherwise, it's a worker with no custom roles.
-	return worker, nil
+	return []*mcfgv1.MachineConfigPool{worker}, nil
+}
+
+// getPrimaryPoolForNode uses getPoolsForNode and returns the first one which is the one the node targets
+func (ctrl *Controller) getPrimaryPoolForNode(node *corev1.Node) (*mcfgv1.MachineConfigPool, error) {
+	pools, err := ctrl.getPoolsForNode(node)
+	if err != nil {
+		return nil, err
+	}
+	if pools == nil {
+		return nil, nil
+	}
+	return pools[0], nil
 }
 
 func (ctrl *Controller) enqueue(pool *mcfgv1.MachineConfigPool) {
@@ -432,7 +675,7 @@ func (ctrl *Controller) syncMachineConfigPool(key string) error {
 		return err
 	}
 
-	if machineconfigpool.Status.Configuration.Name == "" {
+	if machineconfigpool.Spec.Configuration.Name == "" {
 		// Previously we spammed the logs about empty pools.
 		// Let's just pause for a bit here to let the renderer
 		// initialize them.
@@ -447,7 +690,7 @@ func (ctrl *Controller) syncMachineConfigPool(key string) error {
 	everything := metav1.LabelSelector{}
 
 	if reflect.DeepEqual(pool.Spec.NodeSelector, &everything) {
-		ctrl.eventRecorder.Eventf(pool, v1.EventTypeWarning, "SelectingAll", "This machineconfigpool is selecting all nodes. A non-empty selector is required.")
+		ctrl.eventRecorder.Eventf(pool, corev1.EventTypeWarning, "SelectingAll", "This machineconfigpool is selecting all nodes. A non-empty selector is required.")
 		return nil
 	}
 
@@ -459,11 +702,7 @@ func (ctrl *Controller) syncMachineConfigPool(key string) error {
 		return ctrl.syncStatusOnly(pool)
 	}
 
-	selector, err := metav1.LabelSelectorAsSelector(pool.Spec.NodeSelector)
-	if err != nil {
-		return fmt.Errorf("invalid label selector: %v", err)
-	}
-	nodes, err := ctrl.nodeLister.List(selector)
+	nodes, err := ctrl.getNodesForPool(pool)
 	if err != nil {
 		return err
 	}
@@ -471,17 +710,46 @@ func (ctrl *Controller) syncMachineConfigPool(key string) error {
 	maxunavail, err := maxUnavailable(pool, nodes)
 	if err != nil {
 		return err
-	}	
+	}
 
 	candidates := getCandidateMachines(pool, nodes, maxunavail)
 	for _, node := range candidates {
-		if err := ctrl.setDesiredMachineConfigAnnotation(node.Name, pool.Status.Configuration.Name); err != nil {
+		if err := ctrl.setDesiredMachineConfigAnnotation(node.Name, pool.Spec.Configuration.Name); err != nil {
 			return err
 		}
 	}
 	return ctrl.syncStatusOnly(pool)
 }
 
+func (ctrl *Controller) getNodesForPool(pool *mcfgv1.MachineConfigPool) ([]*corev1.Node, error) {
+	selector, err := metav1.LabelSelectorAsSelector(pool.Spec.NodeSelector)
+	if err != nil {
+		return nil, fmt.Errorf("invalid label selector: %v", err)
+	}
+
+	initialNodes, err := ctrl.nodeLister.List(selector)
+	if err != nil {
+		return nil, err
+	}
+
+	nodes := []*corev1.Node{}
+	for _, n := range initialNodes {
+		p, err := ctrl.getPrimaryPoolForNode(n)
+		if err != nil {
+			glog.Warningf("can't get pool for node %q: %v", n.Name, err)
+			continue
+		}
+		if p == nil {
+			continue
+		}
+		if p.Name != pool.Name {
+			continue
+		}
+		nodes = append(nodes, n)
+	}
+	return nodes, nil
+}
+
 func (ctrl *Controller) setDesiredMachineConfigAnnotation(nodeName, currentConfig string) error {
 	glog.Infof("Setting node %s to desired config %s", nodeName, currentConfig)
 	return clientretry.RetryOnConflict(nodeUpdateBackoff, func() error {
@@ -507,7 +775,7 @@ func (ctrl *Controller) setDesiredMachineConfigAnnotation(nodeName, currentConfi
 			return err
 		}
 
-		patchBytes, err := strategicpatch.CreateTwoWayMergePatch(oldData, newData, v1.Node{})
+		patchBytes, err := strategicpatch.CreateTwoWayMergePatch(oldData, newData, corev1.Node{})
 		if err != nil {
 			return fmt.Errorf("failed to create patch for node %q: %v", nodeName, err)
 		}
@@ -517,7 +785,7 @@ func (ctrl *Controller) setDesiredMachineConfigAnnotation(nodeName, currentConfi
 }
 
 func getCandidateMachines(pool *mcfgv1.MachineConfigPool, nodesInPool []*corev1.Node, maxUnavailable int) []*corev1.Node {
-	targetConfig := pool.Status.Configuration.Name
+	targetConfig := pool.Spec.Configuration.Name
 
 	unavail := getUnavailableMachines(nodesInPool)
 	// If we're at capacity, there's nothing to do.
@@ -554,16 +822,10 @@ func getCandidateMachines(pool *mcfgv1.MachineConfigPool, nodesInPool []*corev1.
 }
 
 func maxUnavailable(pool *mcfgv1.MachineConfigPool, nodes []*corev1.Node) (int, error) {
-	// Hardcoded for now per discussion in
-	// https://github.com/openshift/machine-config-operator/pull/701
-	if pool.Name == "master" {
-		return 1, nil
-	}
 	intOrPercent := intstrutil.FromInt(1)
 	if pool.Spec.MaxUnavailable != nil {
 		intOrPercent = *pool.Spec.MaxUnavailable
 	}
-
 	maxunavail, err := intstrutil.GetValueFromIntOrPercent(&intOrPercent, len(nodes), false)
 	if err != nil {
 		return 0, err
@@ -571,5 +833,22 @@ func maxUnavailable(pool *mcfgv1.MachineConfigPool, nodes []*corev1.Node) (int,
 	if maxunavail == 0 {
 		maxunavail = 1
 	}
+	if pool.Name == "master" {
+		// calculate the fault tolerance dynamically for the master pool
+		// to avoid risking losing etcd quorum.
+		tolerance := len(nodes) - ((len(nodes) / 2) + 1)
+		if maxunavail > tolerance {
+			glog.Warningf("Refusing to honor master pool maxUnavailable %d to prevent losing etcd quorum, using %d instead", maxunavail, tolerance)
+			return tolerance, nil
+		}
+	}
 	return maxunavail, nil
 }
+
+// getErrorString returns error string if not nil and empty string if error is nil
+func getErrorString(err error) string {
+	if err != nil {
+		return err.Error()
+	}
+	return ""
+}
