@@ -5,7 +5,6 @@ import (
 	"bytes"
 	"encoding/json"
 	"fmt"
-	"io"
 	"os"
 	"os/exec"
 	"os/user"
@@ -16,17 +15,16 @@ import (
 	"syscall"
 	"time"
 
-	ignv2_2types "github.com/coreos/ignition/config/v2_2/types"
-	"github.com/coreos/ignition/config/validate"
+	igntypes "github.com/coreos/ignition/config/v2_2/types"
 	"github.com/golang/glog"
 	"github.com/google/renameio"
-	drain "github.com/openshift/kubernetes-drain"
+	drain "github.com/openshift/cluster-api/pkg/drain"
 	mcfgv1 "github.com/openshift/machine-config-operator/pkg/apis/machineconfiguration.openshift.io/v1"
+	ctrlcommon "github.com/openshift/machine-config-operator/pkg/controller/common"
 	"github.com/openshift/machine-config-operator/pkg/daemon/constants"
 	errors "github.com/pkg/errors"
 	"github.com/vincent-petithory/dataurl"
 	corev1 "k8s.io/api/core/v1"
-	"k8s.io/apimachinery/pkg/types"
 	"k8s.io/apimachinery/pkg/util/wait"
 )
 
@@ -39,6 +37,8 @@ const (
 	coreUserName = "core"
 	// SSH Keys for user "core" will only be written at /home/core/.ssh
 	coreUserSSHPath = "/home/core/.ssh/"
+	// fipsCommand is the command to use when enabling or disabling FIPS
+	fipsCommand = "/usr/libexec/rhcos-tools/coreos-fips"
 )
 
 func writeFileAtomicallyWithDefaults(fpath string, b []byte) error {
@@ -79,17 +79,30 @@ func getNodeRef(node *corev1.Node) *corev1.ObjectReference {
 	return &corev1.ObjectReference{
 		Kind: "Node",
 		Name: node.GetName(),
-		UID:  types.UID(node.GetUID()),
+		UID:  node.GetUID(),
 	}
 }
 
+type drainLogger struct{}
+
+func (dl *drainLogger) Logf(format string, v ...interface{}) {
+	glog.Infof(format, v...)
+}
+
+func (dl *drainLogger) Log(v ...interface{}) {
+	glog.Info(v...)
+}
+
 // updateOSAndReboot is the last step in an update(), and it can also
 // be called as a special case for the "bootstrap pivot".
 func (dn *Daemon) updateOSAndReboot(newConfig *mcfgv1.MachineConfig) (retErr error) {
 	if err := dn.updateOS(newConfig); err != nil {
 		return err
 	}
+	return dn.drainAndReboot(newConfig)
+}
 
+func (dn *Daemon) drainAndReboot(newConfig *mcfgv1.MachineConfig) (retErr error) {
 	if out, err := dn.storePendingState(newConfig, 1); err != nil {
 		return errors.Wrapf(err, "failed to log pending config: %s", string(out))
 	}
@@ -109,7 +122,7 @@ func (dn *Daemon) updateOSAndReboot(newConfig *mcfgv1.MachineConfig) (retErr err
 	}
 
 	// Skip draining of the node when we're not cluster driven
-	if dn.onceFrom == "" {
+	if dn.kubeClient != nil {
 		dn.logSystem("Update prepared; beginning drain")
 
 		dn.recorder.Eventf(getNodeRef(dn.node), corev1.EventTypeNormal, "Drain", "Draining node to update config.")
@@ -124,8 +137,9 @@ func (dn *Daemon) updateOSAndReboot(newConfig *mcfgv1.MachineConfig) (retErr err
 			err := drain.Drain(dn.kubeClient, []*corev1.Node{dn.node}, &drain.DrainOptions{
 				DeleteLocalData:    true,
 				Force:              true,
-				GracePeriodSeconds: 600,
+				GracePeriodSeconds: -1,
 				IgnoreDaemonsets:   true,
+				Logger:             &drainLogger{},
 			})
 			if err == nil {
 				return true, nil
@@ -143,7 +157,7 @@ func (dn *Daemon) updateOSAndReboot(newConfig *mcfgv1.MachineConfig) (retErr err
 	}
 
 	// reboot. this function shouldn't actually return.
-	return dn.reboot(fmt.Sprintf("Node will reboot into config %v", newConfig.GetName()), defaultRebootTimeout, rebootCommand(fmt.Sprintf("Node will reboot into config %v", newConfig.GetName())))
+	return dn.reboot(fmt.Sprintf("Node will reboot into config %v", newConfig.GetName()))
 }
 
 var errUnreconcilable = errors.New("unreconcilable")
@@ -169,25 +183,35 @@ func (dn *Daemon) update(oldConfig, newConfig *mcfgv1.MachineConfig) (retErr err
 		}
 	}()
 
-	oldConfigName := oldConfig.GetName()
-	newConfigName := newConfig.GetName()
-	glog.Infof("Checking reconcilable for config %v to %v", oldConfigName, newConfigName)
-	// make sure we can actually reconcile this state
-	reconcilableError := dn.reconcilable(oldConfig, newConfig)
+	oldConfigUnset := oldConfig == nil
+	if oldConfigUnset {
+		// Rather than change the rest of the code to deal
+		// with a nil oldConfig, just pass an empty one in
+		// most places.
+		emptyMC := mcfgv1.MachineConfig{}
+		oldConfig = &emptyMC
+	} else {
+		oldConfigName := oldConfig.GetName()
+		newConfigName := newConfig.GetName()
+		glog.Infof("Checking Reconcilable for config %v to %v", oldConfigName, newConfigName)
 
-	if reconcilableError != nil {
-		wrappedErr := fmt.Errorf("can't reconcile config %s with %s: %v", oldConfigName, newConfigName, reconcilableError)
-		if dn.recorder != nil {
-			mcRef := &corev1.ObjectReference{
-				Kind: "MachineConfig",
-				Name: newConfig.GetName(),
-				UID:  newConfig.GetUID(),
+		// make sure we can actually reconcile this state
+		diff, reconcilableError := Reconcilable(oldConfig, newConfig)
+
+		if reconcilableError != nil {
+			wrappedErr := fmt.Errorf("can't reconcile config %s with %s: %v", oldConfigName, newConfigName, reconcilableError)
+			if dn.recorder != nil {
+				mcRef := &corev1.ObjectReference{
+					Kind: "MachineConfig",
+					Name: newConfig.GetName(),
+					UID:  newConfig.GetUID(),
+				}
+				dn.recorder.Eventf(mcRef, corev1.EventTypeWarning, "FailedToReconcile", wrappedErr.Error())
 			}
-			dn.recorder.Eventf(mcRef, corev1.EventTypeWarning, "FailedToReconcile", wrappedErr.Error())
+			return errors.Wrapf(errUnreconcilable, "%v", wrappedErr)
 		}
-		return errors.Wrapf(errUnreconcilable, "%v", wrappedErr)
+		dn.logSystem("Starting update from %s to %s: %+v", oldConfigName, newConfigName, diff)
 	}
-	dn.logSystem("Starting update from %s to %s", oldConfigName, newConfigName)
 
 	// update files on disk that need updating
 	if err := dn.updateFiles(oldConfig, newConfig); err != nil {
@@ -220,7 +244,7 @@ func (dn *Daemon) update(oldConfig, newConfig *mcfgv1.MachineConfig) (retErr err
 		return err
 	}
 	defer func() {
-		if retErr != nil {
+		if retErr != nil && !oldConfigUnset {
 			if err := dn.storeCurrentConfigOnDisk(oldConfig); err != nil {
 				retErr = errors.Wrapf(retErr, "error rolling back current config on disk %v", err)
 				return
@@ -228,10 +252,67 @@ func (dn *Daemon) update(oldConfig, newConfig *mcfgv1.MachineConfig) (retErr err
 		}
 	}()
 
+	// kargs
+	if err := dn.updateKernelArguments(oldConfig, newConfig); err != nil {
+		return err
+	}
+	defer func() {
+		if retErr != nil {
+			if err := dn.updateKernelArguments(newConfig, oldConfig); err != nil {
+				retErr = errors.Wrapf(retErr, "error rolling back kernel arguments %v", err)
+				return
+			}
+		}
+	}()
+
+	if err := dn.updateFIPS(oldConfig, newConfig); err != nil {
+		return err
+	}
+	defer func() {
+		if retErr != nil {
+			if err := dn.updateFIPS(newConfig, oldConfig); err != nil {
+				retErr = errors.Wrapf(retErr, "error rolling back FIPS %v", err)
+				return
+			}
+		}
+	}()
+
 	return dn.updateOSAndReboot(newConfig)
 }
 
-// reconcilable checks the configs to make sure that the only changes requested
+func (dn *Daemon) updateFIPS(current, desired *mcfgv1.MachineConfig) error {
+	if current.Spec.Fips == desired.Spec.Fips {
+		return nil
+	}
+	if dn.OperatingSystem != machineConfigDaemonOSRHCOS {
+		return errors.New("Updating FIPS on non-RHCOS nodes is not supported")
+	}
+	arg := "enable"
+	if !desired.Spec.Fips {
+		arg = "disable"
+	}
+	cmd := exec.Command(fipsCommand, arg)
+	dn.logSystem("Running %s %s", fipsCommand, arg)
+	if out, err := cmd.CombinedOutput(); err != nil {
+		return errors.Wrapf(err, "%s FIPS: %s", arg, string(out))
+	}
+	return nil
+}
+
+// MachineConfigDiff represents an ad-hoc difference between two MachineConfig objects.
+// At some point this may change into holding just the files/units that changed
+// and the MCO would just operate on that.  For now we're just doing this to get
+// improved logging.
+type MachineConfigDiff struct {
+	osUpdate bool
+	kargs    bool
+	fips     bool
+	passwd   bool
+	files    bool
+	units    bool
+}
+
+// Reconcilable checks the configs to make sure that the only changes requested
 // are ones we know how to do in-place.  If we can reconcile, (nil, nil) is returned.
 // Otherwise, if we can't do it in place, the node is marked as degraded;
 // the returned string value includes the rationale.
@@ -239,28 +320,20 @@ func (dn *Daemon) update(oldConfig, newConfig *mcfgv1.MachineConfig) (retErr err
 // we can only update machine configs that have changes to the files,
 // directories, links, and systemd units sections of the included ignition
 // config currently.
-
-func (dn *Daemon) reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) error {
-	// We skip out of reconcilable if there is no Kind and we are in runOnce mode. The
-	// reason is that there is a good chance a previous state is not available to match against.
-	if oldConfig.Kind == "" && dn.onceFrom != "" {
-		glog.Info("Missing kind in old config. Assuming no prior state.")
-		return nil
-	}
+func Reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) (*MachineConfigDiff, error) {
 	oldIgn := oldConfig.Spec.Config
 	newIgn := newConfig.Spec.Config
 
 	// Ignition section
 	// First check if this is a generally valid Ignition Config
-	rpt := validate.ValidateWithoutSource(reflect.ValueOf(newIgn))
-	if rpt.IsFatal() {
-		return errors.Errorf("invalid Ignition config found: %v", rpt)
+	if err := ctrlcommon.ValidateIgnition(newIgn); err != nil {
+		return nil, err
 	}
 
 	// if the config versions are different, all bets are off. this probably
 	// shouldn't happen, but if it does, we can't deal with it.
 	if oldIgn.Ignition.Version != newIgn.Ignition.Version {
-		return fmt.Errorf("ignition version mismatch between old and new config: old: %s new: %s",
+		return nil, fmt.Errorf("ignition version mismatch between old and new config: old: %s new: %s",
 			oldIgn.Ignition.Version, newIgn.Ignition.Version)
 	}
 	// everything else in the ignition section doesn't matter to us, since the
@@ -273,7 +346,7 @@ func (dn *Daemon) reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) error
 	// we don't currently configure the network in place. we can't fix it if
 	// something changed here.
 	if !reflect.DeepEqual(oldIgn.Networkd, newIgn.Networkd) {
-		return errors.New("ignition networkd section contains changes")
+		return nil, errors.New("ignition networkd section contains changes")
 	}
 
 	// Passwd section
@@ -281,24 +354,25 @@ func (dn *Daemon) reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) error
 	// we don't currently configure Groups in place. we don't configure Users except
 	// for setting/updating SSHAuthorizedKeys for the only allowed user "core".
 	// otherwise we can't fix it if something changed here.
-	if !reflect.DeepEqual(oldIgn.Passwd, newIgn.Passwd) {
+	passwdChanged := !reflect.DeepEqual(oldIgn.Passwd, newIgn.Passwd)
+	if passwdChanged {
 		if !reflect.DeepEqual(oldIgn.Passwd.Groups, newIgn.Passwd.Groups) {
-			return errors.New("ignition Passwd Groups section contains changes")
+			return nil, errors.New("ignition Passwd Groups section contains changes")
 		}
 		if !reflect.DeepEqual(oldIgn.Passwd.Users, newIgn.Passwd.Users) {
 			// check if the prior config is empty and that this is the first time running.
 			// if so, the SSHKey from the cluster config and user "core" must be added to machine config.
-			if len(oldIgn.Passwd.Users) >= 0 && len(newIgn.Passwd.Users) >= 1 {
+			if len(oldIgn.Passwd.Users) > 0 && len(newIgn.Passwd.Users) >= 1 {
 				// there is an update to Users, we must verify that it is ONLY making an acceptable
 				// change to the SSHAuthorizedKeys for the user "core"
 				for _, user := range newIgn.Passwd.Users {
 					if user.Name != coreUserName {
-						return errors.New("ignition passwd user section contains unsupported changes: non-core user")
+						return nil, errors.New("ignition passwd user section contains unsupported changes: non-core user")
 					}
 				}
 				glog.Infof("user data to be verified before ssh update: %v", newIgn.Passwd.Users[len(newIgn.Passwd.Users)-1])
 				if err := verifyUserFields(newIgn.Passwd.Users[len(newIgn.Passwd.Users)-1]); err != nil {
-					return err
+					return nil, err
 				}
 			}
 		}
@@ -309,23 +383,23 @@ func (dn *Daemon) reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) error
 	// we can only reconcile files right now. make sure the sections we can't
 	// fix aren't changed.
 	if !reflect.DeepEqual(oldIgn.Storage.Disks, newIgn.Storage.Disks) {
-		return errors.New("ignition disks section contains changes")
+		return nil, errors.New("ignition disks section contains changes")
 	}
 	if !reflect.DeepEqual(oldIgn.Storage.Filesystems, newIgn.Storage.Filesystems) {
-		return errors.New("ignition filesystems section contains changes")
+		return nil, errors.New("ignition filesystems section contains changes")
 	}
 	if !reflect.DeepEqual(oldIgn.Storage.Raid, newIgn.Storage.Raid) {
-		return errors.New("ignition raid section contains changes")
+		return nil, errors.New("ignition raid section contains changes")
 	}
 	if !reflect.DeepEqual(oldIgn.Storage.Directories, newIgn.Storage.Directories) {
-		return errors.New("ignition directories section contains changes")
+		return nil, errors.New("ignition directories section contains changes")
 	}
 	if !reflect.DeepEqual(oldIgn.Storage.Links, newIgn.Storage.Links) {
 		// This means links have been added, as opposed as being removed as it happened with
 		// https://bugzilla.redhat.com/show_bug.cgi?id=1677198. This doesn't really change behavior
 		// since we still don't support links but we allow old MC to remove links when upgrading.
 		if len(newIgn.Storage.Links) != 0 {
-			return errors.New("ignition links section contains changes")
+			return nil, errors.New("ignition links section contains changes")
 		}
 	}
 
@@ -333,7 +407,7 @@ func (dn *Daemon) reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) error
 	// have to force a reprovision since it's not idempotent
 	for _, f := range newIgn.Storage.Files {
 		if f.Append {
-			return fmt.Errorf("ignition file %v includes append", f.Path)
+			return nil, fmt.Errorf("ignition file %v includes append", f.Path)
 		}
 	}
 
@@ -343,7 +417,14 @@ func (dn *Daemon) reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) error
 
 	// we made it through all the checks. reconcile away!
 	glog.V(2).Info("Configs are reconcilable")
-	return nil
+	return &MachineConfigDiff{
+		osUpdate: oldConfig.Spec.OSImageURL != newConfig.Spec.OSImageURL,
+		kargs:    !reflect.DeepEqual(oldConfig.Spec.KernelArguments, newConfig.Spec.KernelArguments),
+		fips:     oldConfig.Spec.Fips != newConfig.Spec.Fips,
+		passwd:   passwdChanged,
+		files:    !reflect.DeepEqual(oldIgn.Storage.Files, newIgn.Storage.Files),
+		units:    !reflect.DeepEqual(oldIgn.Systemd.Units, newIgn.Systemd.Units),
+	}, nil
 }
 
 // verifyUserFields returns nil if the user Name = "core", if 1 or more SSHKeys exist for
@@ -351,8 +432,8 @@ func (dn *Daemon) reconcilable(oldConfig, newConfig *mcfgv1.MachineConfig) error
 // Otherwise, an error will be returned and the proposed config will not be reconcilable.
 // At this time we do not support non-"core" users or any changes to the "core" user
 // outside of SSHAuthorizedKeys.
-func verifyUserFields(pwdUser ignv2_2types.PasswdUser) error {
-	emptyUser := ignv2_2types.PasswdUser{}
+func verifyUserFields(pwdUser igntypes.PasswdUser) error {
+	emptyUser := igntypes.PasswdUser{}
 	tempUser := pwdUser
 	if tempUser.Name == coreUserName && len(tempUser.SSHAuthorizedKeys) >= 1 {
 		tempUser.Name = ""
@@ -367,6 +448,49 @@ func verifyUserFields(pwdUser ignv2_2types.PasswdUser) error {
 	return nil
 }
 
+// generateKargsCommand performs a diff between the old/new MC kernelArguments,
+// and generates the command line arguments suitable for `rpm-ostree kargs`.
+// Note what we really should be doing though is also looking at the *current*
+// kernel arguments in case there was drift.  But doing that requires us knowing
+// what the "base" arguments are.  See https://github.com/ostreedev/ostree/issues/479
+func generateKargsCommand(oldConfig, newConfig *mcfgv1.MachineConfig) []string {
+	oldKargs := make(map[string]bool)
+	for _, arg := range oldConfig.Spec.KernelArguments {
+		oldKargs[arg] = true
+	}
+	newKargs := make(map[string]bool)
+	for _, arg := range newConfig.Spec.KernelArguments {
+		newKargs[arg] = true
+	}
+	cmdArgs := []string{}
+	for _, arg := range oldConfig.Spec.KernelArguments {
+		if !newKargs[arg] {
+			cmdArgs = append(cmdArgs, "--delete="+arg)
+		}
+	}
+	for _, arg := range newConfig.Spec.KernelArguments {
+		if !oldKargs[arg] {
+			cmdArgs = append(cmdArgs, "--append="+arg)
+		}
+	}
+	return cmdArgs
+}
+
+// updateKernelArguments adjusts the kernel args
+func (dn *Daemon) updateKernelArguments(oldConfig, newConfig *mcfgv1.MachineConfig) error {
+	diff := generateKargsCommand(oldConfig, newConfig)
+	if len(diff) == 0 {
+		return nil
+	}
+	if dn.OperatingSystem != machineConfigDaemonOSRHCOS {
+		return fmt.Errorf("Updating kargs on non-RHCOS nodes is not supported: %v", diff)
+	}
+
+	args := append([]string{"kargs"}, diff...)
+	dn.logSystem("Running rpm-ostree %v", args)
+	return exec.Command("rpm-ostree", args...).Run()
+}
+
 // updateFiles writes files specified by the nodeconfig to disk. it also writes
 // systemd units. there is no support for multiple filesystems at this point.
 //
@@ -412,6 +536,13 @@ func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) er
 
 	for _, f := range oldConfig.Spec.Config.Storage.Files {
 		if _, ok := newFileSet[f.Path]; !ok {
+			if _, err := os.Stat(origFileName(f.Path)); err == nil {
+				if err := os.Rename(origFileName(f.Path), f.Path); err != nil {
+					return err
+				}
+				glog.V(2).Infof("Restored file %q", f.Path)
+				continue
+			}
 			glog.V(2).Infof("Deleting stale config file: %s", f.Path)
 			if err := os.Remove(f.Path); err != nil {
 				newErr := fmt.Errorf("unable to delete %s: %s", f.Path, err)
@@ -474,7 +605,7 @@ func (dn *Daemon) deleteStaleData(oldConfig, newConfig *mcfgv1.MachineConfig) er
 }
 
 // enableUnit enables a systemd unit via symlink
-func (dn *Daemon) enableUnit(unit ignv2_2types.Unit) error {
+func (dn *Daemon) enableUnit(unit igntypes.Unit) error {
 	// The link location
 	wantsPath := filepath.Join(wantsPathSystemd, unit.Name)
 	// sanity check that we don't return an error when the link already exists
@@ -486,16 +617,15 @@ func (dn *Daemon) enableUnit(unit ignv2_2types.Unit) error {
 	servicePath := filepath.Join(pathSystemd, unit.Name)
 	err := renameio.Symlink(servicePath, wantsPath)
 	if err != nil {
-		glog.Warningf("Cannot enable unit %s: %s", unit.Name, err)
-	} else {
-		glog.Infof("Enabled %s", unit.Name)
-		glog.V(2).Infof("Symlinked %s to %s", servicePath, wantsPath)
+		return err
 	}
-	return err
+	glog.Infof("Enabled %s", unit.Name)
+	glog.V(2).Infof("Symlinked %s to %s", servicePath, wantsPath)
+	return nil
 }
 
 // disableUnit disables a systemd unit via symlink removal
-func (dn *Daemon) disableUnit(unit ignv2_2types.Unit) error {
+func (dn *Daemon) disableUnit(unit igntypes.Unit) error {
 	// The link location
 	wantsPath := filepath.Join(wantsPathSystemd, unit.Name)
 	// sanity check so we don't return an error when the unit was already disabled
@@ -509,7 +639,7 @@ func (dn *Daemon) disableUnit(unit ignv2_2types.Unit) error {
 }
 
 // writeUnits writes the systemd units to disk
-func (dn *Daemon) writeUnits(units []ignv2_2types.Unit) error {
+func (dn *Daemon) writeUnits(units []igntypes.Unit) error {
 	for _, u := range units {
 		// write the dropin to disk
 		for i := range u.Dropins {
@@ -585,7 +715,7 @@ func (dn *Daemon) writeUnits(units []ignv2_2types.Unit) error {
 
 // writeFiles writes the given files to disk.
 // it doesn't fetch remote files and expects a flattened config file.
-func (dn *Daemon) writeFiles(files []ignv2_2types.File) error {
+func (dn *Daemon) writeFiles(files []igntypes.File) error {
 	for _, file := range files {
 		glog.Infof("Writing file %q", file.Path)
 
@@ -607,6 +737,9 @@ func (dn *Daemon) writeFiles(files []ignv2_2types.File) error {
 				return fmt.Errorf("failed to retrieve file ownership for file %q: %v", file.Path, err)
 			}
 		}
+		if err := createOrigFile(file.Path); err != nil {
+			return err
+		}
 		if err := writeFileAtomically(file.Path, contents.Data, defaultDirectoryPermissions, mode, uid, gid); err != nil {
 			return err
 		}
@@ -614,8 +747,35 @@ func (dn *Daemon) writeFiles(files []ignv2_2types.File) error {
 	return nil
 }
 
+func origParentDir() string {
+	return filepath.Join("/etc", "machine-config-daemon", "orig")
+}
+
+func origFileName(fpath string) string {
+	return filepath.Join(origParentDir(), fpath+".mcdorig")
+}
+
+func createOrigFile(fpath string) error {
+	if _, err := os.Stat(fpath); err != nil {
+		// the file isn't there, no need to back it up
+		// we could check ENOENT only maybe?
+		return nil
+	}
+	if _, err := os.Stat(origFileName(fpath)); err == nil {
+		// the orig file is already there and we avoid creating a new one to preserve the real default
+		return nil
+	}
+	if err := os.MkdirAll(filepath.Dir(origFileName(fpath)), 0755); err != nil {
+		return errors.Wrapf(err, "creating orig parent dir: %v", err)
+	}
+	if out, err := exec.Command("cp", "-a", "--reflink=auto", fpath, origFileName(fpath)).CombinedOutput(); err != nil {
+		return errors.Wrapf(err, "creating orig file for %q: %s", fpath, string(out))
+	}
+	return nil
+}
+
 // This is essentially ResolveNodeUidAndGid() from Ignition; XXX should dedupe
-func getFileOwnership(file ignv2_2types.File) (int, int, error) {
+func getFileOwnership(file igntypes.File) (int, int, error) {
 	uid, gid := 0, 0 // default to root
 	if file.User != nil {
 		if file.User.ID != nil {
@@ -644,7 +804,7 @@ func getFileOwnership(file ignv2_2types.File) (int, int, error) {
 	return uid, gid, nil
 }
 
-func (dn *Daemon) atomicallyWriteSSHKey(newUser ignv2_2types.PasswdUser, keys string) error {
+func (dn *Daemon) atomicallyWriteSSHKey(keys string) error {
 	authKeyPath := filepath.Join(coreUserSSHPath, "authorized_keys")
 
 	// Keys should only be written to "/home/core/.ssh"
@@ -661,7 +821,7 @@ func (dn *Daemon) atomicallyWriteSSHKey(newUser ignv2_2types.PasswdUser, keys st
 }
 
 // Update a given PasswdUser's SSHKey
-func (dn *Daemon) updateSSHKeys(newUsers []ignv2_2types.PasswdUser) error {
+func (dn *Daemon) updateSSHKeys(newUsers []igntypes.PasswdUser) error {
 	if len(newUsers) == 0 {
 		return nil
 	}
@@ -672,9 +832,11 @@ func (dn *Daemon) updateSSHKeys(newUsers []ignv2_2types.PasswdUser) error {
 	for _, k := range newUsers[len(newUsers)-1].SSHAuthorizedKeys {
 		concatSSHKeys = concatSSHKeys + string(k) + "\n"
 	}
-	// newUsers[0] is currently unused since we write keys only for the core user
-	if err := dn.atomicSSHKeysWriter(newUsers[0], concatSSHKeys); err != nil {
-		return err
+	if !dn.mock {
+		// Note we write keys only for the core user and so this ignores the user list
+		if err := dn.atomicallyWriteSSHKey(concatSSHKeys); err != nil {
+			return err
+		}
 	}
 	return nil
 }
@@ -706,29 +868,16 @@ func (dn *Daemon) updateOS(config *mcfgv1.MachineConfig) error {
 	return nil
 }
 
-// RHEL 7.6 logger (util-linux) doesn't have the --journald flag
-func (dn *Daemon) isLoggingToJournalSupported() bool {
-	loggerOutput, err := exec.Command("logger", "--help").CombinedOutput()
-	if err != nil {
-		dn.logSystem("error running logger --help: %v", err)
-		if dn.OperatingSystem == machineConfigDaemonOSRHCOS {
-			return true
-		}
-		return false
-	}
-	return strings.Contains(string(loggerOutput), "--journald")
-}
-
-func (dn *Daemon) getPendingStateLegacyLogger() (string, error) {
+func (dn *Daemon) getPendingStateLegacyLogger() (*journalMsg, error) {
 	glog.Info("logger doesn't support --jounald, grepping the journal")
 
-	cmdLiteral := "journalctl -o cat | grep OPENSHIFT_MACHINE_CONFIG_DAEMON_LEGACY_LOG_HACK"
+	cmdLiteral := "journalctl -o cat _UID=0 | grep OPENSHIFT_MACHINE_CONFIG_DAEMON_LEGACY_LOG_HACK"
 	cmd := exec.Command("bash", "-c", cmdLiteral)
 	var combinedOutput bytes.Buffer
 	cmd.Stdout = &combinedOutput
 	cmd.Stderr = &combinedOutput
 	if err := cmd.Start(); err != nil {
-		return "", errors.Wrap(err, "failed shelling out to journalctl -o cat")
+		return nil, errors.Wrap(err, "failed shelling out to journalctl -o cat")
 	}
 	if err := cmd.Wait(); err != nil {
 		if exiterr, ok := err.(*exec.ExitError); ok {
@@ -738,44 +887,43 @@ func (dn *Daemon) getPendingStateLegacyLogger() (string, error) {
 				// grep exit with 1 if it doesn't find anything
 				// from man: Normally, the exit status is 0 if selected lines are found and 1 otherwise. But the exit status is 2 if an error occurred
 				if status.ExitStatus() == 1 {
-					return "", nil
+					return nil, nil
 				}
 				if status.ExitStatus() > 1 {
-					return "", errors.Wrapf(fmt.Errorf("grep exited with %s", combinedOutput.Bytes()), "failed to grep on journal output: %v", exiterr)
+					return nil, errors.Wrapf(fmt.Errorf("grep exited with %s", combinedOutput.Bytes()), "failed to grep on journal output: %v", exiterr)
 				}
 			}
 		} else {
-			return "", errors.Wrap(err, "command wait error")
+			return nil, errors.Wrap(err, "command wait error")
 		}
 	}
 	journalOutput := combinedOutput.Bytes()
 	// just an extra safety check?
 	if len(journalOutput) == 0 {
-		return "", nil
+		return nil, nil
 	}
 	return dn.processJournalOutput(journalOutput)
 }
 
-func (dn *Daemon) processJournalOutput(journalOutput []byte) (string, error) {
+type journalMsg struct {
+	Message   string `json:"MESSAGE,omitempty"`
+	BootID    string `json:"BOOT_ID,omitempty"`
+	Pending   string `json:"PENDING,omitempty"`
+	OldLogger string `json:"OPENSHIFT_MACHINE_CONFIG_DAEMON_LEGACY_LOG_HACK,omitempty"` // unused today
+}
+
+func (dn *Daemon) processJournalOutput(journalOutput []byte) (*journalMsg, error) {
 	lines := strings.Split(strings.TrimSpace(string(journalOutput)), "\n")
 	last := lines[len(lines)-1]
-	type journalMsg struct {
-		Message   string `json:"MESSAGE,omitempty"`
-		BootID    string `json:"BOOT_ID,omitempty"`
-		Pending   string `json:"PENDING,omitempty"`
-		OldLogger string `json:"OPENSHIFT_MACHINE_CONFIG_DAEMON_LEGACY_LOG_HACK,omitempty"` // unused today
-	}
+
 	entry := &journalMsg{}
 	if err := json.Unmarshal([]byte(last), entry); err != nil {
-		return "", errors.Wrap(err, "getting pending state from journal")
+		return nil, errors.Wrap(err, "getting pending state from journal")
 	}
 	if entry.Pending == "0" {
-		return "", nil
+		return nil, nil
 	}
-	if entry.BootID == dn.bootID {
-		return "", fmt.Errorf("pending config %s bootID %s matches current! Failed to reboot?", entry.Message, dn.bootID)
-	}
-	return entry.Message, nil
+	return entry, nil
 }
 
 // getPendingState loads the JSON state we cache across attempting to apply
@@ -783,16 +931,16 @@ func (dn *Daemon) processJournalOutput(journalOutput []byte) (string, error) {
 // The bootID is stored in the pending state; if it is unchanged, we assume
 // that we failed to reboot; that for now should be a fatal error, in order to avoid
 // reboot loops.
-func (dn *Daemon) getPendingState() (string, error) {
+func (dn *Daemon) getPendingState() (*journalMsg, error) {
 	if !dn.loggerSupportsJournal {
 		return dn.getPendingStateLegacyLogger()
 	}
-	journalOutput, err := exec.Command("journalctl", "-o", "json", fmt.Sprintf("MESSAGE_ID=%s", pendingStateMessageID)).CombinedOutput()
+	journalOutput, err := exec.Command("journalctl", "-o", "json", "_UID=0", fmt.Sprintf("MESSAGE_ID=%s", pendingStateMessageID)).CombinedOutput()
 	if err != nil {
-		return "", errors.Wrap(err, "error running journalctl -o json")
+		return nil, errors.Wrap(err, "error running journalctl -o json")
 	}
 	if len(journalOutput) == 0 {
-		return "", nil
+		return nil, nil
 	}
 	return dn.processJournalOutput(journalOutput)
 }
@@ -821,10 +969,10 @@ func (dn *Daemon) storePendingState(pending *mcfgv1.MachineConfig, isPending int
 	logger := exec.Command("logger", "--journald")
 
 	var pendingState bytes.Buffer
-	pendingState.Write([]byte(fmt.Sprintf(`MESSAGE_ID=%s
+	pendingState.WriteString(fmt.Sprintf(`MESSAGE_ID=%s
 MESSAGE=%s
 BOOT_ID=%s
-PENDING=%d`, pendingStateMessageID, pending.GetName(), dn.bootID, isPending)))
+PENDING=%d`, pendingStateMessageID, pending.GetName(), dn.bootID, isPending))
 
 	logger.Stdin = &pendingState
 	return logger.CombinedOutput()
@@ -839,21 +987,13 @@ func (dn *Daemon) logSystem(format string, a ...interface{}) {
 	// subprocess rather than talking to journald in process since
 	// I worry about the golang library having a connection pre-chroot.
 	logger := exec.Command("logger")
-	stdin, err := logger.StdinPipe()
-	if err != nil {
-		glog.Errorf("failed to get stdin pipe: %v", err)
-		return
-	}
 
-	go func() {
-		defer stdin.Close()
-		io.WriteString(stdin, fmt.Sprintf("machine-config-daemon[%d]: ", os.Getpid()))
-		io.WriteString(stdin, message)
-	}()
-	err = logger.Run()
-	if err != nil {
+	var log bytes.Buffer
+	log.WriteString(fmt.Sprintf("machine-config-daemon[%d]: %s", os.Getpid(), message))
+
+	logger.Stdin = &log
+	if err := logger.Run(); err != nil {
 		glog.Errorf("failed to invoke logger: %v", err)
-		return
 	}
 }
 
@@ -877,22 +1017,22 @@ func (dn *Daemon) cancelSIGTERM() {
 // reboot is the final step. it tells systemd-logind to reboot the machine,
 // cleans up the agent's connections, and then sleeps for 7 days. if it wakes up
 // and manages to return, it returns a scary error message.
-func (dn *Daemon) reboot(rationale string, timeout time.Duration, rebootCmd *exec.Cmd) error {
+func (dn *Daemon) reboot(rationale string) error {
+	// Now that everything is done, avoid delaying shutdown.
+	dn.cancelSIGTERM()
+	dn.Close()
+
+	if dn.skipReboot {
+		return nil
+	}
+
 	// We'll only have a recorder if we're cluster driven
 	if dn.recorder != nil {
 		dn.recorder.Eventf(getNodeRef(dn.node), corev1.EventTypeNormal, "Reboot", rationale)
 	}
 	dn.logSystem("initiating reboot: %s", rationale)
 
-	// Now that everything is done, avoid delaying shutdown.
-	dn.cancelSIGTERM()
-
-	dn.Close()
-
-	if dn.skipReboot && dn.onceFrom != "" { // the dn.onceFrom check is just a triple check that we're not messing with in-cluster MCDs
-		glog.Info("MCD is not rebooting in onceFrom with --skip-reboot")
-		return nil
-	}
+	rebootCmd := rebootCommand(rationale)
 
 	// reboot, executed async via systemd-run so that the reboot command is executed
 	// in the context of the host asynchronously from us
@@ -904,7 +1044,7 @@ func (dn *Daemon) reboot(rationale string, timeout time.Duration, rebootCmd *exe
 	}
 
 	// wait to be killed via SIGTERM from the kubelet shutting down
-	time.Sleep(timeout)
+	time.Sleep(defaultRebootTimeout)
 
 	// if everything went well, this should be unreachable.
 	return fmt.Errorf("reboot failed; this error should be unreachable, something is seriously wrong")
